;; Generated package description from forgejo.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "forgejo" "0.1.4.0.20260504.135909" "Emacs Forgejo Front-end" '((emacs "29.1") (keymap-popup "0.2.1")) :commit "e7e55e6ba2dc50a16987215a438239e2501b7be7" :authors '(("Thanos Apollo" . "public@thanosapollo.org")) :maintainer '("Thanos Apollo" . "public@thanosapollo.org") :keywords '("tools" "vc" "git" "forgejo") :url "https://codeberg.org/thanosapollo/emacs-forgejo")
